#include "null.h"
#include "datashadow.h"

/*
* Multi-valued lookup table implementation
* to store/retrieve/lookup shadow data (addresses)
* using an index as the key value
*
* Implements per-entry lock using gcc builtins for atomic lock operations
*    see lock_shadow_table(index), unlock_shadow_table(index)
*/

/* data shadow table */
static struct data_shadow_entry_manager DataShadowTable[DATA_SHADOW_TABLE_SIZE];

/* preallocate pool of data shadow entries */
static struct data_shadow_entry DataShadowEntryPool[MAX_DATA_SHADOW_ENTRIES];

static int module_initialized = 0;

// #define lock_shadow_table(X) while (__sync_lock_test_and_set(&DataShadowTable[X].locked, 1));
// #define unlock_shadow_table(X) __sync_lock_release(&DataShadowTable[X].locked);

void lock_shadow_table(int p_idx)
{
	// NOP for now
	// (1) start with global lock
	// (2) then continue with fine-grained (per shadow index) lobal lock
}

void unlock_shadow_table(int p_idx)
{
		// NOOP for now
//	__sync_lock_release(&DataShadowTable[p_idx].locked);
}

void initialize_data_shadow_module()
{
	int i;

	if (module_initialized) return;

	// initialize shadow table
	for (i = 0; i < DATA_SHADOW_TABLE_SIZE; ++i)
	{
		DataShadowTable[i].shadow_entry = NO_SHADOW_ENTRY;
	}

	// initialize preallocated shadow data entries
	for (i = 0; i < MAX_DATA_SHADOW_ENTRIES; ++i)
	{
		DataShadowEntryPool[i].val.addr = NULL;
		DataShadowEntryPool[i].next_shadow_entry = NO_SHADOW_ENTRY;
	}

	module_initialized = 1;
}


static int get_next_free_entry()
{
static int next_free_entry = 0;

	if (next_free_entry >= (MAX_DATA_SHADOW_ENTRIES - 1))
		return -1;

	int current_free_entry = next_free_entry;
	next_free_entry++;
	return current_free_entry;
}

static int find_shadow_entry(int p_idx, void *p_addr)
{
	int entry = DataShadowTable[p_idx].shadow_entry;
	while (entry != NO_SHADOW_ENTRY)
	{
		if (DataShadowEntryPool[entry].val.addr == p_addr)
		{
			return 1;
		}

		entry = DataShadowEntryPool[entry].next_shadow_entry;
	}

	return 0;
}

int add_shadow_entry(int p_idx, void *p_addr)
{
	if (!module_initialized)
		initialize_data_shadow_module();
				
	if (p_idx < 0 || p_idx >= DATA_SHADOW_TABLE_SIZE) return 0;

	lock_shadow_table(p_idx);

	if (!find_shadow_entry(p_idx, p_addr))
	{
		int new_entry = get_next_free_entry();
		DataShadowEntryPool[new_entry].val.addr = p_addr;

		if (DataShadowTable[p_idx].shadow_entry == NO_SHADOW_ENTRY)
		{
			// completely new entry
			DataShadowTable[p_idx].shadow_entry = new_entry;
			DataShadowEntryPool[new_entry].next_shadow_entry = NO_SHADOW_ENTRY;
		}
		else
		{
			// insert to front
			SHADOW_ENTRY_INDEX old_entry = DataShadowTable[p_idx].shadow_entry;
			DataShadowTable[p_idx].shadow_entry = new_entry;
			DataShadowEntryPool[new_entry].next_shadow_entry = old_entry;
		}
	}

	unlock_shadow_table(p_idx);

	return 1;
}

/*
* \param p_idx index into global shadow table
* \param p_addr address to check
* \param p_restore_addr (out) set to shadowed value
* \returns 1 if found, 0 if not found
*/
int check_shadow_value(int p_idx, void *p_addr, void **p_restore_addr, int *count)
{
	int entry;
	int matched = 0;
	void *address_visited = NULL;

	if (!module_initialized)
		initialize_data_shadow_module();

	lock_shadow_table(p_idx);

	*p_restore_addr = NULL;
	*count = 0;

	// traverse all entries, find a match
	entry = DataShadowTable[p_idx].shadow_entry;
	while (entry != NO_SHADOW_ENTRY)
	{
		(*count)++;
		address_visited = DataShadowEntryPool[entry].val.addr;
		if (address_visited == p_addr)
		{
			*p_restore_addr = p_addr;
			unlock_shadow_table(p_idx);
			return 1;
		}
		entry = DataShadowEntryPool[entry].next_shadow_entry;
	}
	
	//
	// if we get here, we found no matching address
	//    ==> <p_addr> is invalid
	//

	if (*count == 0)
	{
		// no entry in the table, this is an error
		*p_restore_addr = NULL;
	}
	else if (*count == 1)
	{
		// common case: only 1 entry -- set to shadowed value
		*p_restore_addr = address_visited;
	}
	else
	{
		// multiple entries -- what to do?
		// for now return one of the entries
		*p_restore_addr = address_visited;
	}

	unlock_shadow_table(p_idx);

	return matched;
}
